from time import perf_counter

VERSION = (2, 10, 0)
__version__ = '.'.join(map(str, VERSION))

# Export pyperf.perf_counter for backward compatibility with pyperf 1.7
# which supports Python 2 and Python 3
__all__ = ['perf_counter']

from pyperf._utils import python_implementation, python_has_jit  # noqa
__all__.extend(('python_implementation', 'python_has_jit'))

from pyperf._metadata import format_metadata  # noqa
__all__.append('format_metadata')

from pyperf._bench import Run, Benchmark, BenchmarkSuite, add_runs  # noqa
__all__.extend(('Run', 'Benchmark', 'BenchmarkSuite', 'add_runs'))

from pyperf._runner import Runner   # noqa
__all__.append('Runner')



import os
try:
    from pyperf._utils import USE_PSUTIL, BSD
    if not USE_PSUTIL:
        raise ImportError
    else:
        import psutil
except ImportError:
    raise ImportError('psutil is not installed')
import threading
import time


class PeakMemoryUsageThread(threading.Thread):
    def __init__(self):
        threading.Thread.__init__(self)
        self.process = psutil.Process(os.getpid())
        self.peak_usage = 0
        self._done = threading.Event()
        self.sleep = 0.010   # 10 ms
        self._quit = False

    def get(self):
        if BSD:
            # USS (Unique Set Size) is not supported on BSD,
            # use RSS (Resident Set Size) instead.
            usage = self.process.memory_info().rss
        else:
            usage = self.process.memory_full_info().uss
        self.peak_usage = max(self.peak_usage, usage)

    def run(self):
        try:
            while not self._quit:
                self.get()
                time.sleep(self.sleep)
        finally:
            self._done.set()

    def stop(self):
        self._quit = True
        self._done.wait()
        return self.peak_usage


def check_tracking_memory():
    mem_thread = PeakMemoryUsageThread()
    mem_thread.get()

    if not mem_thread.peak_usage:
        return "memory usage is zero"

    return None



import unittest

import pyperf
from pyperf import _cli as cli


class CLITests(unittest.TestCase):
    def test_format_result(self):
        run = pyperf.Run([1.0, 1.5, 2.0],
                         warmups=[(1, 3.0)],
                         metadata={'name': 'mybench'},
                         collect_metadata=False)
        bench = pyperf.Benchmark([run])
        self.assertEqual(cli.format_result_value(bench),
                         '1.50 sec +- 0.50 sec')
        self.assertEqual(cli.format_result(bench),
                         'Mean +- std dev: 1.50 sec +- 0.50 sec')

    def test_format_result_calibration(self):
        run = pyperf.Run([], warmups=[(100, 1.0)],
                         metadata={'name': 'bench', 'loops': 100},
                         collect_metadata=False)
        bench = pyperf.Benchmark([run])
        self.assertEqual(cli.format_result_value(bench),
                         '<calibration: 100 loops>')
        self.assertEqual(cli.format_result(bench),
                         'Calibration: 100 loops')
        self.assertRaises(ValueError, bench.median)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
import pyperf

runner = pyperf.Runner()
runner.timeit("[1]*1000",
              stmt="[1]*1000",
              duplicate=1024)
runner.timeit("[1,2]*1000",
              stmt="[1,2]*1000",
              duplicate=1024)
runner.timeit("[1,2,3]*1000",
              stmt="[1,2,3]*1000",
              duplicate=1024)



from pyperf._cli import display_title, format_result_value
from pyperf._utils import is_significant, geometric_mean


def is_significant_benchs(bench1, bench2):
    values1 = bench1.get_values()
    values2 = bench2.get_values()

    if len(values1) == 1 and len(values2) == 1:
        # FIXME: is it ok to consider that comparison between two values
        # is significant?
        return (True, None)

    try:
        significant, t_score = is_significant(values1, values2)
        return (significant, t_score)
    except Exception:
        # FIXME: fix the root bug, don't work around it
        return (True, None)


class CompareData:
    def __init__(self, name, benchmark):
        self.name = name
        self.benchmark = benchmark

    def __repr__(self):
        return '<CompareData name=%r value#=%s>' % (self.name, self.benchmark.get_nvalue())


def compute_normalized_mean(bench, ref):
    ref_avg = ref.mean()
    bench_avg = bench.mean()
    # Note: means cannot be zero, it's a warranty of pyperf API
    return bench_avg / ref_avg


def format_normalized_mean(norm_mean):
    if norm_mean == 1.0:
        return "no change"
    elif norm_mean < 1.0:
        return "%.2fx faster" % (1.0 / norm_mean)
    else:
        return "%.2fx slower" % norm_mean


def format_geometric_mean(norm_means):
    geo_mean = geometric_mean(norm_means)
    return format_normalized_mean(geo_mean)


def get_tags_for_result(result):
    return result.ref.benchmark.get_metadata().get("tags", [])


class CompareResult:
    def __init__(self, ref, changed, min_speed=None):
        # CompareData object
        self.ref = ref
        # CompareData object
        self.changed = changed
        self._min_speed = min_speed
        self._significant = None
        self._t_score = None
        self._norm_mean = None

    def __repr__(self):
        return '<CompareResult ref=%r changed=%r>' % (self.ref, self.changed)

    def _set_significant(self):
        bench1 = self.ref.benchmark
        bench2 = self.changed.benchmark
        self._significant, self._t_score = is_significant_benchs(bench1, bench2)

        if self._min_speed is not None:
            norm_mean = self.norm_mean
            if norm_mean < 1.0:
                # faster uses the inverse
                norm_mean = 1.0 / norm_mean
            if (norm_mean - 1.0) * 100 < self._min_speed:
                self._significant = False

    @property
    def significant(self):
        if self._significant is None:
            self._set_significant()
        return self._significant

    @property
    def t_score(self):
        if self._significant is None:
            self._set_significant()
        return self._t_score

    def _compute_norm_mean(self):
        ref = self.ref.benchmark
        bench = self.changed.benchmark
        self._norm_mean = compute_normalized_mean(bench, ref)

    # mean normalized to the reference benchmark mean
    @property
    def norm_mean(self):
        if self._norm_mean is None:
            self._compute_norm_mean()
        return self._norm_mean

    def oneliner(self, verbose=True, show_name=True, check_significant=True):
        if check_significant and not self.significant:
            return "Not significant!"

        ref_text = format_result_value(self.ref.benchmark)
        chg_text = format_result_value(self.changed.benchmark)
        if verbose:
            if show_name:
                ref_text = "[%s] %s" % (self.ref.name, ref_text)
                chg_text = "[%s] %s" % (self.changed.name, chg_text)
            if (self.ref.benchmark.get_nvalue() > 1
               or self.changed.benchmark.get_nvalue() > 1):
                text = "Mean +- std dev: %s -> %s" % (ref_text, chg_text)
            else:
                text = "%s -> %s" % (ref_text, chg_text)
        else:
            text = "%s -> %s" % (ref_text, chg_text)

        text = "%s: %s" % (text, format_normalized_mean(self.norm_mean))
        return text

    def format(self, verbose=True, show_name=True):
        text = self.oneliner(show_name=show_name, check_significant=False)
        lines = [text]

        # significant?
        if self.t_score is None:
            lines.append("ERROR when testing if values are significant")

        if self.significant:
            if verbose:
                if self.t_score is not None:
                    lines.append("Significant (t=%.2f)" % self.t_score)
                else:
                    lines.append("Significant")
        else:
            lines.append("Not significant!")
        return lines


class CompareResults(list):
    # list of CompareResult objects
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return '<CompareResult %r>' % (list(self),)


class ReSTTable:
    def __init__(self, headers, rows):
        self.headers = headers
        self.rows = rows
        self.widths = [len(header) for header in self.headers]
        for row in self.rows:
            for column, cell in enumerate(row):
                self.widths[column] = max(self.widths[column], len(cell))

    def _render_line(self, char='-'):
        parts = ['']
        for width in self.widths:
            parts.append(char * (width + 2))
        parts.append('')
        return '+'.join(parts)

    def _render_row(self, row):
        parts = ['']
        for width, cell in zip(self.widths, row):
            parts.append(' %s ' % cell.ljust(width))
        parts.append('')
        return '|'.join(parts)

    def render(self, write_line):
        write_line(self._render_line('-'))
        write_line(self._render_row(self.headers))
        write_line(self._render_line('='))
        for row in self.rows:
            write_line(self._render_row(row))
            write_line(self._render_line('-'))


class MarkDownTable:
    def __init__(self, headers, rows):
        self.headers = headers
        self.rows = rows
        self.widths = [len(header) for header in self.headers]
        for row in self.rows:
            for column, cell in enumerate(row):
                self.widths[column] = max(self.widths[column], len(cell))

    def _render_line(self, char='-'):
        parts = ['']
        for idx, width in enumerate(self.widths):
            if idx == 0:
                parts.append(char * (width + 2))
            else:
                parts.append(f':{char * width}:')
        parts.append('')
        return '|'.join(parts)

    def _render_row(self, row):
        parts = ['']
        for width, cell in zip(self.widths, row):
            parts.append(" %s " % cell.ljust(width))
        parts.append('')
        return '|'.join(parts)

    def render(self, write_line):
        write_line(self._render_row(self.headers))
        write_line(self._render_line('-'))
        for row in self.rows:
            write_line(self._render_row(row))


class CompareError(Exception):
    pass


class CompareSuites:
    def __init__(self, benchmarks, args):
        self.benchmarks = benchmarks

        self.table = args.table
        self.table_format = args.table_format
        self.min_speed = args.min_speed
        self.group_by_speed = args.group_by_speed
        self.verbose = args.verbose
        self.quiet = args.quiet

        grouped_by_name = self.benchmarks.group_by_name()
        if not grouped_by_name:
            raise CompareError("Benchmark suites have no benchmark in common")

        # List of CompareResults
        self.all_results = []
        for item in grouped_by_name:
            cmp_benchmarks = item.benchmarks
            results = self.compare_benchmarks(item.name, cmp_benchmarks)
            self.all_results.append(results)

        self.show_name = (len(grouped_by_name) > 1)

        self.tags = set()
        for results in self.all_results:
            for result in results:
                self.tags.update(get_tags_for_result(result))
        self.tags = sorted(list(self.tags))

    def compare_benchmarks(self, name, benchmarks):
        min_speed = self.min_speed

        results = CompareResults(name)

        ref_item = benchmarks[0]
        ref = CompareData(ref_item.filename, ref_item.benchmark)

        for item in benchmarks[1:]:
            changed = CompareData(item.filename, item.benchmark)
            result = CompareResult(ref, changed, min_speed)
            results.append(result)

        return results

    @staticmethod
    def display_not_significant(not_significant):
        print("Benchmark hidden because not significant (%s): %s"
              % (len(not_significant), ', '.join(not_significant)))

    def compare_suites_table(self, all_results):
        if self.group_by_speed:
            def sort_key(results):
                result = results[0]
                return result.norm_mean

            self.all_results.sort(key=sort_key)

        headers = ['Benchmark', self.all_results[0][0].ref.name]
        for item in self.all_results[0]:
            headers.append(item.changed.name)

        all_norm_means = [[] for _ in range(len(headers[2:]))]

        rows = []
        not_significant = []
        for results in all_results:
            row = [results.name]

            ref_bench = results[0].ref.benchmark
            text = ref_bench.format_value(ref_bench.mean())
            row.append(text)

            significants = []
            for index, result in enumerate(results):
                bench = result.changed.benchmark
                significant = result.significant
                if significant:
                    text = format_normalized_mean(result.norm_mean)
                    if not self.quiet:
                        text = "%s: %s" % (bench.format_value(bench.mean()), text)
                else:
                    text = "not significant"
                significants.append(significant)
                all_norm_means[index].append(result.norm_mean)
                row.append(text)

            if any(significants):
                rows.append(row)
            else:
                not_significant.append(results.name)

        # only compute the geometric mean if there is at least two benchmarks
        # and if at least one is signicant.
        if len(all_norm_means[0]) > 1 and rows:
            row = ['Geometric mean', '(ref)']
            for norm_means in all_norm_means:
                row.append(format_geometric_mean(norm_means))
            rows.append(row)

        if rows:
            if self.table_format == 'rest':
                table = ReSTTable(headers, rows)
            else:
                table = MarkDownTable(headers, rows)
            table.render(print)

        if not_significant:
            if rows:
                print()
            self.display_not_significant(not_significant)

    def compare_suites_by_speed(self, all_results):
        not_significant = []
        slower = []
        faster = []
        same = []
        for results in all_results:
            result = results[0]
            if not result.significant:
                not_significant.append(results.name)
                continue

            item = (results.name, result)
            norm_mean = result.norm_mean
            if norm_mean == 1.0:
                same.append(item)
            elif norm_mean < 1.0:
                faster.append(item)
            else:
                slower.append(item)

        def sort_key(item):
            return item[1].norm_mean

        slower.sort(key=sort_key, reverse=True)
        faster.sort(key=sort_key)

        empty_line = False
        for title, results, sort_reverse in (
            ('Slower', slower, True),
            ('Faster', faster, False),
            ('Same speed', same, False),
        ):
            if not results:
                continue

            if empty_line:
                print()
            print("%s (%s):" % (title, len(results)))
            for name, result in results:
                text = result.oneliner(verbose=False)
                print("- %s: %s" % (name, text))
            empty_line = True

        if not self.quiet and not_significant:
            if empty_line:
                print()
            self.display_not_significant(not_significant)

    def compare_suites_list(self, all_results):
        not_significant = []
        empty_line = False
        last_index = (len(self.all_results) - 1)

        for index, results in enumerate(all_results):
            significant = any(result.significant for result in results)
            lines = []
            for result in results:
                lines.extend(result.format(self.verbose))

            if not (significant or self.verbose):
                not_significant.append(results.name)
                continue

            if len(lines) != 1:
                if self.show_name:
                    display_title(results.name)
                for line in lines:
                    print(line)
                if index != last_index:
                    print()
            else:
                text = lines[0]
                if self.show_name:
                    text = '%s: %s' % (results.name, text)
                print(text)
            empty_line = True

        if not self.quiet and not_significant:
            if empty_line:
                print()
            self.display_not_significant(not_significant)

    def list_ignored(self):
        for suite, hidden in self.benchmarks.group_by_name_ignored():
            if not hidden:
                continue
            hidden_names = [bench.get_name() for bench in hidden]
            print("Ignored benchmarks (%s) of %s: %s"
                  % (len(hidden), suite.filename, ', '.join(sorted(hidden_names))))

    def compare_geometric_mean(self, all_results):
        # use a list since two filenames can be identical,
        # even if results are different
        all_norm_means = []
        for item in all_results[0]:
            all_norm_means.append((item.changed.name, []))

        for results in all_results:
            for index, result in enumerate(results):
                all_norm_means[index][1].append(result.norm_mean)

        if len(all_norm_means[0][1]) < 2:
            # only compute the geometric mean when there is at least two benchmarks
            return

        print()
        if len(all_norm_means) > 1:
            display_title('Geometric mean')
            for name, norm_means in all_norm_means:
                geo_mean = format_geometric_mean(norm_means)
                print(f'{name}: {geo_mean}')
        else:
            geo_mean = format_geometric_mean(all_norm_means[0][1])
            print(f'Geometric mean: {geo_mean}')

    def compare_suites(self, results):
        if self.table:
            self.compare_suites_table(results)
        else:
            if self.group_by_speed:
                self.compare_suites_by_speed(results)
            else:
                self.compare_suites_list(results)

            self.compare_geometric_mean(results)

    def compare(self):
        if len(self.tags):
            for tag in self.tags:
                display_title(f"Benchmarks with tag '{tag}':")
                all_results = [
                    results for results in self.all_results
                    if tag is None or tag in get_tags_for_result(results[0])
                ]
                self.compare_suites(all_results)
                print()
            display_title("All benchmarks:")
        self.compare_suites(self.all_results)

        if not self.quiet:
            self.list_ignored()


def compare_suites(benchmarks, args):
    CompareSuites(benchmarks, args).compare()


def timeit_compare_benchs(name1, bench1, name2, bench2, args):
    data1 = CompareData(name1, bench1)
    data2 = CompareData(name2, bench2)
    compare = CompareResult(data1, data2)
    if not args.quiet:
        lines = compare.format(verbose=args.verbose)
        for line in lines:
            print(line)
    else:
        line = compare.oneliner()
        print(line)



"""
Similar to UNIX time command: measure the execution time of a command.

Minimum Python script spawning a program, wait until it completes, and then
write the elapsed time into stdout. Time is measured by the time.perf_counter()
timer.

Python subprocess.Popen() is implemented with fork()+exec(). Minimize the
Python imports to reduce the memory footprint, to reduce the cost of
fork()+exec().

Measure wall-time, not CPU time.

If resource.getrusage() is available: compute the maximum RSS memory in bytes
per process and writes it into stdout as a second line.
"""
import contextlib
import json
import os
import subprocess
import sys
import tempfile
import time

try:
    import resource
except ImportError:
    resource = None


def get_max_rss(*, children):
    if resource is not None:
        if children:
            resource_type = resource.RUSAGE_CHILDREN
        else:
            resource_type = resource.RUSAGE_SELF
        usage = resource.getrusage(resource_type)
        if sys.platform == 'darwin':
            return usage.ru_maxrss
        return usage.ru_maxrss * 1024
    else:
        return 0


def merge_profile_stats_files(src, dst):
    """
    Merging one existing pstats file into another.
    """
    import pstats
    if os.path.isfile(dst):
        src_stats = pstats.Stats(src)
        dst_stats = pstats.Stats(dst)
        dst_stats.add(src_stats)
        dst_stats.dump_stats(dst)
        os.unlink(src)
    else:
        os.rename(src, dst)


def bench_process(loops, args, kw, profile_filename=None):
    max_rss = 0
    range_it = range(loops)
    start_time = time.perf_counter()

    if profile_filename:
        temp_profile_filename = tempfile.mktemp()
        args = [args[0], "-m", "cProfile", "-o", temp_profile_filename] + args[1:]

    for _ in range_it:
        start_rss = get_max_rss(children=True)

        proc = subprocess.Popen(args, **kw)
        with proc:
            proc.wait()

        exitcode = proc.returncode
        if exitcode != 0:
            print("Command failed with exit code %s" % exitcode,
                  file=sys.stderr)
            if profile_filename:
                os.unlink(temp_profile_filename)
            sys.exit(exitcode)

        rss = get_max_rss(children=True) - start_rss
        max_rss = max(max_rss, rss)

        if profile_filename:
            merge_profile_stats_files(
                temp_profile_filename, profile_filename
            )

    dt = time.perf_counter() - start_time
    return (dt, max_rss)


def load_hooks(metadata):
    hook_names = []
    while "--hook" in sys.argv:
        hook_idx = sys.argv.index("--hook")
        hook_name = sys.argv[hook_idx + 1]
        hook_names.append(hook_name)
        del sys.argv[hook_idx]
        del sys.argv[hook_idx]

    if len(hook_names):
        # Only import pyperf if we know we have hooks
        import pyperf._hooks

        hook_managers = pyperf._hooks.instantiate_selected_hooks(hook_names)
        metadata["hooks"] = ", ".join(hook_managers.keys())
    else:
        hook_managers = {}

    return hook_managers


def write_data(dt, max_rss, metadata, out=sys.stdout):
    # Write the data that is communicated back to the main orchestration process.
    # It is three lines containing:
    #    - The runtime (in seconds)
    #    - max_rss (or -1, if not able to compute)
    #    - The metadata to add to the benchmark entry, as a JSON dictionary
    print(dt, file=out)
    print(max_rss or -1, file=out)
    json.dump(metadata, fp=out)
    print(file=out)


def main():
    # Make sure that the pyperf module wasn't imported
    if 'pyperf' in sys.modules:
        print("ERROR: don't run %s -m pyperf._process, run the .py script"
              % os.path.basename(sys.executable))
        sys.exit(1)

    if len(sys.argv) < 3:
        print("Usage: %s %s loops program [arg1 arg2 ...] [--profile profile]"
              % (os.path.basename(sys.executable), __file__))
        sys.exit(1)

    if "--profile" in sys.argv:
        profile_idx = sys.argv.index("--profile")
        profile_filename = sys.argv[profile_idx + 1]
        del sys.argv[profile_idx]
        del sys.argv[profile_idx]
    else:
        profile_filename = None

    metadata = {}
    hook_managers = load_hooks(metadata)

    loops = int(sys.argv[1])
    args = sys.argv[2:]

    kw = {}
    if hasattr(subprocess, 'DEVNULL'):
        devnull = None
        kw['stdin'] = subprocess.DEVNULL
        kw['stdout'] = subprocess.DEVNULL
    else:
        devnull = open(os.devnull, 'w+', 0)
        kw['stdin'] = devnull
        kw['stdout'] = devnull
    kw['stderr'] = subprocess.STDOUT

    with contextlib.ExitStack() as stack:
        for hook in hook_managers.values():
            stack.enter_context(hook)
        dt, max_rss = bench_process(loops, args, kw, profile_filename)

    if devnull is not None:
        devnull.close()

    for hook in hook_managers.values():
        hook.teardown(metadata)

    write_data(dt, max_rss, metadata)


if __name__ == "__main__":
    main()



# Hooks are installable context managers defined as entry points so that
# arbitrary code can by run right before and after the actual internal
# benchmarking code is run.


import abc
import importlib.metadata
import os.path
import shlex
import subprocess
import sys
import tempfile
import uuid


def get_hooks():
    hook_prefix = "pyperf.hook"
    entry_points = importlib.metadata.entry_points()
    if sys.version_info[:2] < (3, 10):
        group = entry_points[hook_prefix]
    else:
        group = entry_points.select(group=hook_prefix)
    return group


def get_hook_names():
    return (x.name for x in get_hooks())


def get_selected_hooks(hook_names):
    if hook_names is None:
        return

    hook_mapping = {hook.name: hook for hook in get_hooks()}
    for hook_name in hook_names:
        yield hook_mapping[hook_name]


def instantiate_selected_hooks(hook_names):
    hook_managers = {}
    for hook in get_selected_hooks(hook_names):
        try:
            hook_managers[hook.name] = hook.load()()
        except HookError as e:
            print(f"ERROR setting up hook '{hook.name}':", file=sys.stderr)
            print(str(e), file=sys.stderr)
            sys.exit(1)

    return hook_managers


class HookError(Exception):
    pass


class HookBase(abc.ABC):
    def __init__(self):
        """
        Create a new instance of the hook.
        """
        pass

    def teardown(self, _metadata):
        """
        Called when the hook is completed for a process. May add any information
        collected to the passed-in `metadata` dictionary.
        """
        pass

    def __enter__(self):
        """
        Called immediately before running benchmark code.

        May be called multiple times per instance.
        """
        pass

    def __exit__(self, _exc_type, _exc_value, _traceback):
        """
        Called immediately after running benchmark code.
        """
        pass


class _test_hook(HookBase):
    def __init__(self):
        self._count = 0

    def teardown(self, metadata):
        metadata["_test_hook"] = self._count

    def __enter__(self):
        self._count += 1

    def __exit__(self, _exc_type, _exc_value, _traceback):
        pass


class pystats(HookBase):
    def __init__(self):
        if not hasattr(sys, "_stats_on"):
            raise HookError(
                "Can not collect pystats because python was not built with --enable-pystats"
            )
        sys._stats_off()
        sys._stats_clear()

    def teardown(self, metadata):
        metadata["pystats"] = "enabled"

    def __enter__(self):
        sys._stats_on()

    def __exit__(self, _exc_type, _exc_value, _traceback):
        sys._stats_off()


class perf_record(HookBase):
    """Profile the benchmark using perf-record.

    Profile data is written to the current directory directory by default, or
    to the value of the `PYPERF_PERF_RECORD_DATA_DIR` environment variable, if
    it is provided.

    Profile data files have a basename of the form `perf.data.<uuid>`

    The value of the `PYPERF_PERF_RECORD_EXTRA_OPTS` environment variable is
    appended to the command line of perf-record, if provided.
    """

    def __init__(self):
        self.tempdir = tempfile.TemporaryDirectory()
        self.ctl_fifo = self.mkfifo(self.tempdir.name, "ctl_fifo")
        self.ack_fifo = self.mkfifo(self.tempdir.name, "ack_fifo")
        perf_data_dir = os.environ.get("PYPERF_PERF_RECORD_DATA_DIR", "")
        perf_data_basename = f"perf.data.{uuid.uuid4()}"
        cmd = ["perf", "record",
               "--pid", str(os.getpid()),
               "--output", os.path.join(perf_data_dir, perf_data_basename),
               "--control", f"fifo:{self.ctl_fifo},{self.ack_fifo}"]
        extra_opts = os.environ.get("PYPERF_PERF_RECORD_EXTRA_OPTS", "")
        cmd += shlex.split(extra_opts)
        self.perf = subprocess.Popen(
            cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        self.ctl_fd = open(self.ctl_fifo, "w")
        self.ack_fd = open(self.ack_fifo, "r")

    def __enter__(self):
        self.exec_perf_cmd("enable")

    def __exit__(self, _exc_type, _exc_value, _traceback):
        self.exec_perf_cmd("disable")

    def teardown(self, metadata):
        try:
            self.exec_perf_cmd("stop")
            self.perf.wait(timeout=120)
        finally:
            self.ctl_fd.close()
            self.ack_fd.close()

    def mkfifo(self, tmpdir, basename):
        path = os.path.join(tmpdir, basename)
        os.mkfifo(path)
        return path

    def exec_perf_cmd(self, cmd):
        self.ctl_fd.write(f"{cmd}\n")
        self.ctl_fd.flush()
        self.ack_fd.readline()



import datetime
import errno
import math
import os.path
import sys

import statistics

from pyperf._metadata import (NUMBER_TYPES, parse_metadata,
                              _common_metadata, get_metadata_info,
                              _exclude_common_metadata)
from pyperf._formatter import DEFAULT_UNIT, format_values
from pyperf._utils import median_abs_dev, percentile


# JSON format history:
#
# '1.0' - (pyperf 1.0) warmup values are now per loop iteration,
#         rather than raw values.
# 6 - (pyperf 0.9.6) add common_metadata to the root: metadata common to all
#     benchmarks (common to all runs of all benchmarks); rename 'samples'
#     to 'values' in runs
# 5 - (pyperf 0.8.3) timestamps in metadata are now formatted using a space
#      separator
# 4 - (pyperf 0.7.4) warmups are now a lists of (loops, raw_sample)
#     rather than lists of samples
# 3 - (pyperf 0.7) add Run class
# 2 - (pyperf 0.6) support multiple benchmarks per file
# 1 - first version
_JSON_VERSION = '1.0'
_JSON_MAP_VERSION = {5: (0, 8, 3), 6: (0, 9, 6), '1.0': (1, 0)}

# Metadata checked by add_run(): all runs have must have the same
# value for these metadata (or no run must have this metadata)
_CHECKED_METADATA = (
    'aslr',
    'cpu_count',
    'cpu_model_name',
    'hostname',
    'inner_loops',
    'name',
    'platform',
    'python_executable',
    'python_implementation',
    'python_unicode',
    'python_version',
    'unit')


_UNSET = object()


def _check_warmups(warmups):
    for item in warmups:
        if not isinstance(item, tuple):
            return False
        if len(item) != 2:
            return False

        loops, value = item
        if not isinstance(loops, int):
            return False
        if loops < 1:
            return False

        if not isinstance(value, NUMBER_TYPES):
            return False
        if value < 0:
            return False

    return True


def _cached_attr(func):
    attr = '_' + func.__name__

    def method(self):
        value = getattr(self, attr)
        if value is not None:
            return value

        value = func(self)
        setattr(self, attr, value)
        return value

    return method


class Run:
    # Run is immutable, so it can be shared/exchanged between two benchmarks

    __slots__ = ('_warmups', '_values', '_metadata')

    def __init__(self, values, warmups=None,
                 metadata=None, collect_metadata=True):
        if any(not (isinstance(value, NUMBER_TYPES) and value > 0)
               for value in values):
            raise ValueError("values must be a sequence of number > 0.0")

        if warmups is not None and not _check_warmups(warmups):
            raise ValueError("warmups must be a sequence of (loops, value) "
                             "where loops is a int >= 1 and value "
                             "is a float >= 0.0")

        # tuple of (loops: int, value) items
        if warmups:
            self._warmups = tuple(warmups)
        else:
            self._warmups = None
        self._values = tuple(values)

        if not self._values and not self._warmups:
            raise ValueError("values and warmups are empty sequence")

        if collect_metadata:
            from pyperf._collect_metadata import collect_metadata as collect_func

            metadata2 = collect_func()
            if metadata is not None:
                metadata2.update(metadata)
                metadata = metadata2
            else:
                metadata = metadata2

        # Metadata dictionary
        if metadata:
            self._metadata = parse_metadata(metadata)
        else:
            self._metadata = {}

    def _replace(self, values=None, warmups=True, metadata=None):
        if values is None:
            values = self._values
        if warmups:
            warmups = self._warmups
        else:
            warmups = None
        if metadata is None:
            # share metadata dict since Run metadata is immutable
            metadata = self._metadata
        run = Run(values, warmups=warmups, collect_metadata=False)
        run._metadata = metadata
        return run

    def _is_calibration(self):
        # Run used to calibrate or recalibration the number of loops,
        # or to calibrate the number of warmups
        return (not self._values)

    def _is_calibration_loops(self):
        if not self._is_calibration():
            return False
        if self._has_metadata('calibrate_loops'):
            return True
        # backward compatibility with pyperf 1.1 and older
        return not any(self._has_metadata(name)
                       for name in ('recalibrate_loops', 'calibrate_warmups',
                                    'recalibrate_warmups'))

    def _is_recalibration_loops(self):
        return self._is_calibration() and self._has_metadata('recalibrate_loops')

    def _is_calibration_warmups(self):
        return self._is_calibration() and self._has_metadata('calibrate_warmups')

    def _is_recalibration_warmups(self):
        return self._is_calibration() and self._has_metadata('recalibrate_warmups')

    def _has_metadata(self, name):
        return (name in self._metadata)

    def _get_calibration_loops(self):
        metadata = self._metadata
        if 'calibrate_loops' in metadata:
            return metadata['calibrate_loops']
        if 'recalibrate_loops' in metadata:
            return metadata['recalibrate_loops']

        if self._is_calibration_loops():
            # backward compatibility with pyperf 1.1 and older
            return self.get_loops()

        raise ValueError("run is not a loop calibration")

    def _get_calibration_warmups(self):
        metadata = self._metadata
        if 'calibrate_warmups' in metadata:
            return metadata['calibrate_warmups']
        if 'recalibrate_warmups' in metadata:
            return metadata['recalibrate_warmups']
        raise ValueError("run is not a warmup calibration")

    def _get_name(self):
        return self._metadata.get('name', None)

    def get_metadata(self):
        return dict(self._metadata)

    @property
    def warmups(self):
        if self._warmups:
            return self._warmups
        else:
            return ()

    @property
    def values(self):
        return self._values

    def get_loops(self):
        return self._metadata.get('loops', 1)

    def get_inner_loops(self):
        return self._metadata.get('inner_loops', 1)

    def get_total_loops(self):
        return self.get_loops() * self.get_inner_loops()

    def _get_raw_values(self, warmups=False):
        raw_values = []

        if warmups and self._warmups:
            inner_loops = self.get_inner_loops()
            raw_values.extend(value * (loops * inner_loops)
                              for loops, value in self._warmups)

        total_loops = self.get_total_loops()
        raw_values.extend(value * total_loops for value in self._values)
        return raw_values

    def _remove_warmups(self):
        if not self._warmups:
            return self

        return self._replace(warmups=False)

    def _get_duration(self):
        duration = self._metadata.get('duration', None)
        if duration is not None:
            return duration
        raw_values = self._get_raw_values(warmups=True)
        return math.fsum(raw_values)

    def _get_date(self):
        return self._metadata.get('date', None)

    def _as_json(self, common_metadata):
        data = {}
        if self._warmups:
            data['warmups'] = self._warmups
        if self._values:
            data['values'] = self._values

        metadata = _exclude_common_metadata(self._metadata, common_metadata)
        if metadata:
            data['metadata'] = metadata
        return data

    @classmethod
    def _json_load(cls, version, run_data, common_metadata):
        metadata = run_data.get('metadata', {})
        if common_metadata:
            metadata = dict(common_metadata, **metadata)

        warmups = run_data.get('warmups', None)
        if warmups:
            if version >= (1, 0):
                warmups = [tuple(item) for item in warmups]
            else:
                inner_loops = metadata.get('inner_loops', 1)
                warmups = [(loops, raw_value / (loops * inner_loops))
                           for loops, raw_value in warmups]
        if version >= (0, 9, 6):
            values = run_data.get('values', ())
        else:
            values = run_data['samples']

        return cls(values,
                   warmups=warmups,
                   metadata=metadata,
                   collect_metadata=False)

    def _extract_metadata(self, name):
        value = self._metadata.get(name, None)
        if value is None:
            raise KeyError("run has no metadata %r" % name)

        info = get_metadata_info(name)
        if info.unit:
            metadata = dict(self._metadata, unit=info.unit)
        else:
            metadata = None

        if not isinstance(value, NUMBER_TYPES):
            raise TypeError("metadata %r value is not an integer: got %s"
                            % (name, type(value).__name__))

        return self._replace(values=(value,), warmups=False, metadata=metadata)

    def _remove_all_metadata(self):
        name = self._metadata.get('name', None)
        unit = self._metadata.get('unit', None)
        metadata = {}
        if name:
            metadata['name'] = name
        if unit:
            metadata['unit'] = unit
        return self._replace(metadata=metadata)

    def _update_metadata(self, metadata):
        if 'inner_loops' in metadata:
            inner_loops = self._metadata.get('inner_loops', None)
            if (inner_loops is not None
               and metadata['inner_loops'] != inner_loops):
                raise ValueError("inner_loops metadata cannot be modified")

        metadata2 = dict(self._metadata)
        metadata2.update(metadata)
        return self._replace(metadata=metadata2)


class Benchmark:
    def __init__(self, runs):
        self._runs = []   # list of Run objects
        self._clear_runs_cache()

        if not runs:
            raise ValueError("runs must be a non-empty sequence of Run objects")

        # A benchmark must have a name
        if not runs[0]._has_metadata('name'):
            raise ValueError("A benchmark must have a name: "
                             "the first run has no name metadata")

        for run in runs:
            self.add_run(run)

    def __repr__(self):
        return ('<Benchmark %r with %s runs>'
                % (self.get_name(), len(self._runs)))

    def get_name(self):
        run = self._runs[0]
        return run._get_name()

    def _get_common_metadata(self):
        if self._common_metadata is None:
            runs_metadata = [run._metadata for run in self._runs]
            self._common_metadata = _common_metadata(runs_metadata)
        return self._common_metadata

    def get_metadata(self):
        return dict(self._get_common_metadata())

    def get_total_duration(self):
        durations = [run._get_duration() for run in self._runs]
        return math.fsum(durations)

    def _get_run_property(self, get_property):
        # ignore calibration runs
        values = [get_property(run) for run in self._runs
                  if not run._is_calibration()]
        if len(set(values)) == 1:
            return values[0]

        # Compute the mean (float)
        return math.fsum(values) / len(values)

    def _get_nwarmup(self):
        return self._get_run_property(lambda run: len(run.warmups))

    def _get_nvalue_per_run(self):
        return self._get_run_property(lambda run: len(run.values))

    def get_loops(self):
        return self._get_run_property(lambda run: run.get_loops())

    def get_inner_loops(self):
        return self._get_run_property(lambda run: run.get_inner_loops())

    def get_total_loops(self):
        return self._get_run_property(lambda run: run.get_total_loops())

    def _clear_runs_cache(self, keep_common_metadata=False):
        self._values = None
        self._mean = None
        self._stdev = None
        self._median = None
        self._median_abs_dev = None
        if not keep_common_metadata:
            self._common_metadata = None
        self._dates = _UNSET

    @_cached_attr
    def mean(self):
        value = statistics.mean(self.get_values())
        # add_run() ensures that all values are greater than zero
        if value <= 0:
            raise ValueError("mean must be > 0")
        return value

    @_cached_attr
    def stdev(self):
        values = self.get_values()
        value = statistics.stdev(values)
        # add_run() ensures that all values are greater than zero
        if value < 0:
            raise ValueError("std dev must be >= 0")
        return value

    @_cached_attr
    def median(self):
        value = statistics.median(self.get_values())
        # add_run() ensures that all values are greater than zero
        if value <= 0:
            raise ValueError("median must be > 0")
        return value

    @_cached_attr
    def median_abs_dev(self):
        value = median_abs_dev(self.get_values())
        # add_run() ensures that all values are greater than zero
        if value < 0:
            raise ValueError("MAD must be >= 0")
        return value

    def required_nprocesses(self):
        """
        Determines the number of separate process runs that would be required
        achieve stable results. Specifically, the target is to have 95%
        certainty that there is a variance of less than 1%. If the result is
        greater than the number of processes recorded in the input data, the
        value is meaningless and only means "more samples are required".

        The method used is described in this Wikipedia article about estimating
        the sampling of a mean:

        https://en.wikipedia.org/wiki/Sample_size_determination#Estimation_of_a_mean
        """
        # Get the means of the values per process. The values within the process
        # often vary considerably (e.g. due to cache effects), but the variances
        # between processes should be fairly consistent. Additionally, this
        # value is intended to be advice for the number of processes to run.
        values = []
        for run in self._runs:
            if len(run.values):
                values.append(statistics.mean(run.values))

        if len(values) < 2:
            return None

        total = math.fsum(values)
        mean = total / len(values)
        stddev = statistics.stdev(values)

        # Normalize the stddev so we can target "percentage changed" rather than
        # absolute time
        sigma = stddev / mean

        # 95% certainty
        Z = 1.96
        # 1% variation
        W = 0.01

        # (4Z²σ²)/(W²)
        return math.ceil((4 * Z ** 2 * sigma ** 2) / (W ** 2))

    def percentile(self, p):
        if not (0 <= p <= 100):
            raise ValueError("p must be in the range [0; 100]")
        return percentile(self.get_values(), p / 100.0)

    def add_run(self, run):
        if not isinstance(run, Run):
            raise TypeError("Run expected, got %s" % type(run).__name__)

        # Don't check metadata for the first run
        if self._runs:
            metadata = self._get_common_metadata()
            run_metata = run._metadata
            for key in _CHECKED_METADATA:
                value = metadata.get(key, None)
                run_value = run_metata.get(key, None)
                if run_value != value:
                    raise ValueError("incompatible benchmark, metadata %s is "
                                     "different: current=%s, run=%s"
                                     % (key, value, run_value))

        if self._common_metadata is not None:
            # Update common metadata
            for name, value in list(self._common_metadata.items()):
                if run._metadata.get(name, None) != value:
                    del self._common_metadata[name]
        self._clear_runs_cache(keep_common_metadata=True)

        self._runs.append(run)

    def get_unit(self):
        run = self._runs[0]
        return run._metadata.get('unit', DEFAULT_UNIT)

    def format_values(self, values):
        unit = self.get_unit()
        return format_values(unit, values)

    def format_value(self, value):
        return self.format_values((value,))[0]

    def get_nrun(self):
        return len(self._runs)

    def get_runs(self):
        return list(self._runs)

    def get_nvalue(self):
        if self._values is not None:
            return len(self._values)
        else:
            return sum(len(run.values) for run in self._runs)

    def get_values(self):
        if self._values is not None:
            return self._values

        values = []
        for run in self._runs:
            values.extend(run.values)
        values = tuple(values)
        self._values = values
        return values

    def _get_raw_values(self, warmups=False):
        raw_values = []
        for run in self._runs:
            raw_values.extend(run._get_raw_values(warmups))
        return raw_values

    def _only_calibration(self):
        return all(run._is_calibration() for run in self._runs)

    @classmethod
    def _json_load(cls, version, data, suite_metadata):
        if version >= (0, 9, 6):
            metadata = data.get('metadata', {})
        else:
            metadata = data.get('common_metadata', {})
        metadata = parse_metadata(metadata)
        if suite_metadata:
            metadata = dict(suite_metadata, **metadata)

        runs = []
        for run_data in data['runs']:
            run = Run._json_load(version, run_data, metadata)
            # Don't call add_run() to avoid O(n) complexity:
            # expect that runs were already validated before being written
            # into a JSON file
            runs.append(run)

        return cls(runs)

    def _as_json(self, suite_metadata):
        metadata = self._get_common_metadata()
        common_metadata = dict(metadata, **suite_metadata)

        data = {'runs': [run._as_json(common_metadata) for run in self._runs]}
        metadata = _exclude_common_metadata(metadata, suite_metadata)
        if metadata:
            data['metadata'] = metadata
        return data

    @staticmethod
    def load(file):
        suite = BenchmarkSuite.load(file)
        benchmarks = suite.get_benchmarks()
        if len(benchmarks) != 1:
            raise ValueError("expected 1 benchmark, got %s" % len(benchmarks))
        return benchmarks[0]

    @staticmethod
    def loads(string):
        suite = BenchmarkSuite.loads(string)
        benchmarks = suite.get_benchmarks()
        if len(benchmarks) != 1:
            raise ValueError("expected 1 benchmark, got %s" % len(benchmarks))
        return benchmarks[0]

    def dump(self, file, compact=True, replace=False):
        suite = BenchmarkSuite([self])
        suite.dump(file, compact=compact, replace=replace)

    def _replace_runs(self, new_runs):
        if not new_runs:
            raise ValueError("no more runs")
        self._runs[:] = new_runs
        self._clear_runs_cache()

    def _filter_runs(self, include, only_runs):
        if include:
            old_runs = self._runs
            max_index = len(old_runs) - 1
            runs = []
            for index in only_runs:
                if index <= max_index:
                    runs.append(old_runs[index])
        else:
            runs = self._runs
            max_index = len(runs) - 1
            for index in reversed(only_runs):
                if index <= max_index:
                    del runs[index]
        self._replace_runs(runs)

    def _remove_warmups(self):
        new_runs = [run._remove_warmups() for run in self._runs]
        self._replace_runs(new_runs)

    def add_runs(self, benchmark):
        if not isinstance(benchmark, Benchmark):
            raise TypeError("expected Benchmark, got %s"
                            % type(benchmark).__name__)

        if benchmark is self:
            raise ValueError("cannot add a benchmark to itself")

        for run in benchmark._runs:
            self.add_run(run)

    def get_dates(self):
        if self._dates is not _UNSET:
            return self._dates

        start = None
        end = None
        for run in self._runs:
            run_start = run._get_date()
            if run_start is None:
                continue
            run_start = datetime.datetime.fromisoformat(run_start)

            duration = run._get_duration()
            duration = int(math.ceil(duration))
            run_end = run_start + datetime.timedelta(seconds=duration)
            if start is None or run_start < start:
                start = run_start
            if end is None or run_end > end:
                end = run_end

        if start is not None:
            self._dates = (start, end)
        else:
            self._dates = None
        return self._dates

    def _extract_metadata(self, name):
        new_runs = [run._extract_metadata(name) for run in self._runs]
        self._replace_runs(new_runs)

    def _remove_all_metadata(self):
        new_runs = [run._remove_all_metadata() for run in self._runs]
        self._replace_runs(new_runs)

    def update_metadata(self, metadata):
        metadata = parse_metadata(metadata)
        if not metadata:
            return self

        new_runs = [run._update_metadata(metadata) for run in self._runs]
        self._replace_runs(new_runs)


class BenchmarkSuite:
    def __init__(self, benchmarks, filename=None):
        if not benchmarks:
            raise ValueError("benchmarks must be a non-empty "
                             "sequence of Benchmark objects")

        self.filename = filename
        self._benchmarks = []
        for benchmark in benchmarks:
            self.add_benchmark(benchmark)

    def get_benchmark_names(self):
        return [bench.get_name() for bench in self]

    def get_metadata(self):
        benchs_metadata = [bench._get_common_metadata()
                           for bench in self._benchmarks]
        return _common_metadata(benchs_metadata)

    def __len__(self):
        return len(self._benchmarks)

    def __iter__(self):
        return iter(self._benchmarks)

    def _add_benchmark_runs(self, benchmark):
        name = benchmark.get_name()
        try:
            existing = self.get_benchmark(name)
        except KeyError:
            self.add_benchmark(benchmark)
        else:
            existing.add_runs(benchmark)

    def add_runs(self, result):
        if isinstance(result, Benchmark):
            self._add_benchmark_runs(result)
        elif isinstance(result, BenchmarkSuite):
            for benchmark in result:
                self._add_benchmark_runs(benchmark)
        else:
            raise TypeError("expect Benchmark or BenchmarkSuite, got %s"
                            % type(result).__name__)

    def get_benchmark(self, name):
        for bench in self._benchmarks:
            if bench.get_name() == name:
                return bench
        raise KeyError("there is no benchmark called %r" % name)

    def get_benchmarks(self):
        return list(self._benchmarks)

    def add_benchmark(self, benchmark):
        if benchmark in self._benchmarks:
            raise ValueError("benchmark already part of the suite")

        name = benchmark.get_name()
        if name:
            try:
                self.get_benchmark(name)
            except KeyError:
                pass
            else:
                raise ValueError("the suite has already a benchmark called %r"
                                 % name)

        self._benchmarks.append(benchmark)

    @classmethod
    def _json_load(cls, filename, data):
        version = data.get('version')
        version_info = _JSON_MAP_VERSION.get(version)
        if not version_info:
            raise ValueError("file format version %r not supported" % version)
        benchmarks_json = data['benchmarks']

        if version_info >= (0, 9, 6):
            metadata = data.get('metadata', {})
            if metadata is not None:
                metadata = parse_metadata(metadata)
        else:
            metadata = {}

        benchmarks = []
        for bench_data in benchmarks_json:
            benchmark = Benchmark._json_load(version_info, bench_data, metadata)
            benchmarks.append(benchmark)
        suite = cls(benchmarks, filename=filename)

        if not suite:
            raise ValueError("the file doesn't contain any benchmark")

        return suite

    @staticmethod
    def _load_open(filename):
        if isinstance(filename, bytes):
            suffix = b'.gz'
        else:
            suffix = '.gz'

        if filename.endswith(suffix):
            # Use lazy import to limit imports on 'import pyperf'
            import gzip
            return gzip.open(filename, "rt", encoding="utf-8")
        else:
            return open(filename, "r", encoding="utf-8")

    @classmethod
    def load(cls, file):
        # Use lazy import to limit imports on 'import pyperf'
        import json

        if isinstance(file, (bytes, str)):
            if file != '-':
                filename = file
                fp = cls._load_open(filename)
                with fp:
                    data = json.load(fp)
            else:
                filename = '<stdin>'
                data = json.load(sys.stdin)
        else:
            # file is a file object
            filename = getattr(file, 'name', None)
            data = json.load(file)

        return cls._json_load(filename, data)

    @classmethod
    def loads(cls, string):
        # Use lazy import to limit imports on 'import pyperf'
        import json

        data = json.loads(string)
        return cls._json_load(None, data)

    @staticmethod
    def _dump_open(filename, replace):
        if isinstance(filename, bytes):
            suffix = b'.gz'
        else:
            suffix = '.gz'

        if not replace and os.path.exists(filename):
            raise OSError(errno.EEXIST, "File already exists")

        if filename.endswith(suffix):
            # Use lazy import to limit imports on 'import pyperf'
            import gzip

            return gzip.open(filename, mode="wt", encoding="utf-8")
        else:
            return open(filename, "w", encoding="utf-8")

    def _as_json(self):
        metadata = self.get_metadata()
        benchmarks = [benchmark._as_json(metadata)
                      for benchmark in self._benchmarks]
        data = {'version': _JSON_VERSION, 'benchmarks': benchmarks}
        if metadata:
            data['metadata'] = metadata
        return data

    def dump(self, file, compact=True, replace=False):
        # Use lazy import to limit imports on 'import pyperf'
        import json

        data = self._as_json()

        def dump(data, fp, compact):
            kw = {}
            if compact:
                kw['separators'] = (',', ':')
            else:
                kw['indent'] = 4
            json.dump(data, fp, sort_keys=True, **kw)
            fp.write("\n")
            fp.flush()

        if isinstance(file, (bytes, str)):
            fp = self._dump_open(file, replace)
            with fp:
                dump(data, fp, compact)
                fp.close()
        else:
            # file is a file object
            dump(data, file, compact)

    def _replace_benchmarks(self, benchmarks):
        if not benchmarks:
            raise ValueError("empty benchmark suite")
        self._benchmarks[:] = benchmarks

    def _convert_include_benchmark(self, names):
        name_set = set(names)
        benchmarks = [bench for bench in self
                      if bench.get_name() in name_set]
        if not benchmarks:
            raise KeyError("no benchmark found with name in %r" % names)
        self._replace_benchmarks(benchmarks)

    def _convert_exclude_benchmark(self, names):
        name_set = set(names)
        benchmarks = [bench for bench in self
                      if bench.get_name() not in name_set]
        self._replace_benchmarks(benchmarks)

    def get_total_duration(self):
        durations = [benchmark.get_total_duration() for benchmark in self]
        return math.fsum(durations)

    def get_dates(self):
        start = None
        end = None
        for benchmark in self:
            dates = benchmark.get_dates()
            if not dates:
                continue
            if start is None or dates[0] < start:
                start = dates[0]
            if end is None or dates[1] > end:
                end = dates[1]
        if start is not None:
            return (start, end)
        else:
            return None


def add_runs(filename, result):
    if os.path.exists(filename):
        suite = BenchmarkSuite.load(filename)
        suite.add_runs(result)
        suite.dump(filename, replace=True)
    else:
        result.dump(filename)


def _load_suite_from_pipe(bench_json):
    lines = bench_json.split("\n")
    result = None
    for line in lines:
        if not line:
            continue
        suite = BenchmarkSuite.loads(line)
        if result is not None:
            for bench in suite:
                result.add_benchmark(bench)
        else:
            result = suite
    return result
